#include "codegen/LinearScan.hpp"
#include "codegen/MachineInstruction.hpp"
#include "codegen/MachineOperand.hpp"
#include "codegen/MachineUnit.hpp"
#include "codegen/utils/NotImplemented.hpp"
#include "Function.h"
#include "magic_enum.hpp"
#include "codegen/utils/UtilFunctions.hpp"

#include <iostream>
#include <iomanip>
#include <ostream>
#include <string_view>
#include <vector>

void LinearScan::pass(MachineUnit *mUnit) {

    for (auto mFunc : mUnit->func_list) {
        
        // 按基本块的拓扑序给指令编号
        auto IR_func = mFunc->IR_func;
        for (auto bb : IR_func->getReversePostOrder()) {
            bb->mBlock->begin_no = MachineInstruction::counter;
            for (auto inst : bb->mBlock->inst_list) {
                inst->no = MachineInstruction::counter++;
                mFunc->no2inst[inst->no] = inst;
            }
            bb->mBlock->end_no = MachineInstruction::counter - 1;
        }

        computeLiveIntervals(mFunc);

        
        // 初始化可分配的寄存器
        free_regs = std::move(std::vector<int>(init_free_regs.rbegin(), init_free_regs.rend()));
        // 其他初始化
        intervals.clear(), active.clear();
        // 寄存器分配
        linearScanRegisterAllocation(mFunc);

        rewrite(mFunc);
    }
}

void LinearScan::computeLiveIntervals(MachineFunction *mFunc) {
    auto IR_func = mFunc->IR_func;
    std::vector<BasicBlock *> bb_topo = std::move(IR_func->getReversePostOrder());

    for (auto bb_it = bb_topo.rbegin(); bb_it != bb_topo.rend(); ++bb_it) {
        auto mBlock = (*bb_it)->mBlock;
        for (auto vreg : mBlock->live_out) {
            auto &live_range = mFunc->live_ranges[vreg];
            live_range.insert({mBlock->begin_no, mBlock->end_no});
        }

        for (auto inst_it = mBlock->inst_list.rbegin(); inst_it != mBlock->inst_list.rend(); ++inst_it) {
            for (auto def : (*inst_it)->def_list) {
                if (def->type != MachineOperand::VREG) continue;
                auto &live_range = mFunc->live_ranges[def->vreg_name];

                // if (def->vreg_name == "%4" || def->vreg_name == "%6") {
                //     std::cerr << def->vreg_name <<" :";
                //     for (auto &[l, r] : live_range) std::cerr << "[" << l << ", " << r << "] ";
                //     std::cerr << std::endl;
                // }
                for (auto range_it = live_range.begin(); range_it != live_range.end();) {
                    if ((*inst_it)->no != mBlock->begin_no && range_it->first == mBlock->begin_no) {
                        live_range.insert({(*inst_it)->no, range_it->second});
                        live_range.erase(range_it++);
                    } else {
                        ++range_it;
                    }
                }
                // if (def->vreg_name == "%4" || def->vreg_name == "%6") {
                //     std::cerr << def->vreg_name <<" :";
                //     for (auto &[l, r] : live_range) std::cerr << "[" << l << ", " << r << "] ";
                //     std::cerr << std::endl << std::endl;
                // }
            }

            for (auto use : (*inst_it)->use_list) {
                if (use->type != MachineOperand::VREG) continue;
                auto &live_range = mFunc->live_ranges[use->vreg_name];
                live_range.insert({mBlock->begin_no, (*inst_it)->no});
                if (use->vreg_name == "%4" || use->vreg_name == "%6") {
                    std::cerr << use->vreg_name << "  " << mBlock->begin_no << ' ' << (*inst_it)->no << std::endl;
                }
            }
        }
    }

    for (auto &[vreg, live_range] : mFunc->live_ranges) {
        auto &live_intervals = mFunc->live_intervals;
        live_intervals[vreg] = *live_range.begin();
        for (auto [begin, end] : live_range) {
            live_intervals[vreg].first = std::min(live_intervals[vreg].first, begin);
            live_intervals[vreg].second = std::max(live_intervals[vreg].second, end);
        }
    }
}

std::ostream & operator << (std::ostream &os, const LinearScan::Interval &i) {
    os << setw(6) << i.vreg_name << "   [" << setw(2) << i.begin << ", " << setw(2) << i.end << "]   ";
    if (!i.spilled) {
        os << codegen_utils::lower(magic_enum::enum_name(static_cast<IREGISTER>(i.phy_reg)));
    } else {
        os << i.offset;
    }
    return os;
}

void LinearScan::linearScanRegisterAllocation(MachineFunction *mFunc) {
    for (auto [vreg, intvl] : mFunc->live_intervals) {
        intervals.push_back(new Interval{vreg, intvl.first, intvl.second});
    }

    // 按活跃区间interval的开始点从小到大排序
    std::sort(intervals.begin(), intervals.end(), IntervalBeginCmper());

    for (auto interval : intervals) {
        expireOldIntervals(interval);
        if (active.size() == 11 + 7) {
            spillInterval(interval);
        } else {
            interval->phy_reg = static_cast<IREGISTER>(free_regs.back());
            free_regs.pop_back();
            active.insert(interval);
        }
    }

    for (auto interval : intervals) reg_allocation[interval->vreg_name] = interval;

    std::cerr << "Live Intervals and Allocated Registers of Function '" << mFunc->IR_func->name << "':" << std::endl;
    for (auto interval : intervals) std::cerr << *interval << '\n';
    std::cerr << std::endl;
}

void LinearScan::expireOldIntervals(Interval *i) {
    for (auto it = active.begin(); it != active.end();) {
        auto j = *it;
        if (j->end >= i->begin) return;
        active.erase(it++);
        free_regs.push_back(j->phy_reg);
    }
}

void LinearScan::spillInterval(Interval *i) {
    // (3 temp + 1 slot) * 8 bytes
    int offset = -32;
    auto spill = *active.rbegin();
    if (spill->end > i->end) {
        i->phy_reg = spill->phy_reg;
        spill->spilled = true;
        spill->offset = offset;
        offset -= 8;
        active.erase(spill);
        active.insert(i);
    } else {
        i->spilled = true;
        i->offset = offset;
        offset -= 8;
    }
}

void LinearScan::updateUsedPhysicalRegs(const std::string &func_name, IREGISTER reg) {
    if (T0 <= reg && reg <= T2 || T3 <= reg && reg <= T6) caller_saved_regs[func_name].insert(reg);
    else if (reg == S1 || S2 <= reg && reg <= S11) callee_saved_regs[func_name].insert(reg);
    else if (reg == FP) {
        std::cerr << "对S0的错误使用!";
        assert(false);
    } else {
        std::cerr << "使用了非通用寄存器!";
        assert(false);
    }
}

void LinearScan::rewrite(MachineFunction *mFunc) {
    for (auto [no, inst] : mFunc->no2inst) {
        for (auto &def : inst->def_list) {
            if (def->type != MachineOperand::VREG) continue;
            auto alloc = reg_allocation[def->vreg_name];
            if (!alloc->spilled) {
                updateUsedPhysicalRegs(mFunc->IR_func->name, alloc->phy_reg);
                def = new MachineOperand(MachineOperand::REG, alloc->phy_reg);
            } else {
                // TODO
                throw NotImplemented("寄存器溢出！");
            }
        }

        for (auto &use : inst->use_list) {
            if (use->type != MachineOperand::VREG) continue;
            auto alloc = reg_allocation[use->vreg_name];
            if (!alloc->spilled) {
                updateUsedPhysicalRegs(mFunc->IR_func->name, alloc->phy_reg);
                use = new MachineOperand(MachineOperand::REG, alloc->phy_reg);
            } else {
                // TODO
                throw NotImplemented("寄存器溢出！");
            }
        }
    }

    // 保存 caller-saved registers
    for (auto [no, inst] : mFunc->no2inst) {
        if (inst->type == MachineInstruction::JUMP && inst->optype == JalMInstruction::JAL) {
            // TODO 
            assert(typeid(*inst) == typeid(JalMInstruction));
            auto &label = inst->use_list[1]->label;
            auto &regs = caller_saved_regs[mFunc->IR_func->name];
            if (!regs.empty()) {
                auto addi1 = new BinaryMInstruction(inst->parent, BinaryMInstruction::ADDI,
                                                new MachineOperand(MachineOperand::REG, IREGISTER::SP), 
                                                new MachineOperand(MachineOperand::REG, IREGISTER::SP), 
                                                new MachineOperand(MachineOperand::IMM, -8 * regs.size())
                                            );
                inst->insertBefore(addi1);
                int cnt = 0;
                for (auto reg : regs) {
                    auto sd = new StoreMInstruction(inst->parent,
                                                            StoreMInstruction::SD,
                                                            new MachineOperand(MachineOperand::REG, reg),
                                                            new MachineOperand(MachineOperand::IMM, cnt * 8),
                                                            new MachineOperand(MachineOperand::REG, IREGISTER::SP)
                                                        );
                    inst->insertBefore(sd);
                    ++cnt;
                }
                
                cnt = 0;
                MachineInstruction *cur = inst;
                for (auto reg : regs) {
                    auto ld = new LoadMInstruction(inst->parent,
                                                            LoadMInstruction::LD,
                                                            new MachineOperand(MachineOperand::REG, reg),
                                                            new MachineOperand(MachineOperand::IMM, cnt * 8),
                                                            new MachineOperand(MachineOperand::REG, IREGISTER::SP)
                                                        );
                    cur->insertAfter(ld);
                    cur = ld;
                    ++cnt;
                }

                auto addi2 = new BinaryMInstruction(inst->parent, BinaryMInstruction::ADDI,
                                                new MachineOperand(MachineOperand::REG, IREGISTER::SP), 
                                                new MachineOperand(MachineOperand::REG, IREGISTER::SP), 
                                                new MachineOperand(MachineOperand::IMM, 8 * regs.size())
                                            );
                cur->insertAfter(addi2);
            }
        }
    }
    
    // 保存 callee-saved registers
    auto &regs = callee_saved_regs[mFunc->IR_func->name];
    if (!regs.empty()) {
        throw NotImplemented("保存callee-saved registers");
        // int cnt{0};
        // auto cur = mFunc->no2inst[0];
        // assert(mFunc->block_list[0]), assert(cur->parent == mFunc->block_list[0]), assert(cur->optype == BinaryMInstruction::ADDI);
        // mFunc->framesize -= regs.size() * 8;
        // cur->use_list[1]->val -= regs.size() * 8;

        // cur = mFunc->no2inst[3];
        // cur->use_list[1]->val += regs.size() * 8;

        // for (auto reg : regs) {
        //     auto sd = new StoreMInstruction(cur->parent,
        //                                             StoreMInstruction::SD,
        //                                             new MachineOperand(MachineOperand::REG, reg),
        //                                             new MachineOperand(MachineOperand::IMM, cnt * 8),
        //                                             new MachineOperand(MachineOperand::REG, IREGISTER::SP)
        //                                             );
        //     cur->insertAfter(sd), cur = sd;
        //     ++cnt;
        // }
    }
}